@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -8,6 +13,7 @@
  * Various useful functions for the CVS support code.
  */
 
+#include <assert.h>
 #include "cvs.h"
 #include "getline.h"
 
@@ -51,7 +57,7 @@ xmalloc (bytes)
 /*
  * realloc data and die if it fails [I've always wanted to have "realloc" do
  * a "malloc" if the argument is NULL, but you can't depend on it.  Here, I
- * can *force* it.
+ * can *force* it.]
  */
 void *
 xrealloc (ptr, bytes)
@@ -149,61 +155,70 @@ xstrdup (str)
     return (s);
 }
 
-/* Remove trailing newlines from STRING, destructively. */
-void
-strip_trailing_newlines (str)
-     char *str;
-{
-    int len;
-    len = strlen (str) - 1;
 
-    while (str[len] == '\n')
-	str[len--] = '\0';
+
+/* Remove trailing newlines from STRING, destructively.
+ *
+ * RETURNS
+ *
+ *   True if any newlines were removed, false otherwise.
+ */
+int
+strip_trailing_newlines (str)
+    char *str;
+{
+    size_t index, origlen;
+    index = origlen = strlen (str);
+
+    while (index > 0 && str[index-1] == '\n')
+	str[--index] = '\0';
+
+    return index != origlen;
 }
 
-/* Return the number of levels that path ascends above where it starts.
-   For example:
-   "../../foo" -> 2
-   "foo/../../bar" -> 1
-   */
-/* FIXME: Should be using ISDIRSEP, last_component, or some other
-   mechanism which is more general than just looking at slashes,
-   particularly for the client.c caller.  The server.c caller might
-   want something different, so be careful.  */
+
+
+/* Return the number of levels that PATH ascends above where it starts.
+ * For example:
+ *
+ *   "../../foo" -> 2
+ *   "foo/../../bar" -> 1
+ */
 int
-pathname_levels (path)
-    char *path;
+pathname_levels (p)
+    const char *p;
 {
-    char *p;
-    char *q;
     int level;
     int max_level;
 
+    if (p == NULL) return 0;
+
     max_level = 0;
-    p = path;
     level = 0;
     do
     {
-	q = strchr (p, '/');
-	if (q != NULL)
-	    ++q;
-	if (p[0] == '.' && p[1] == '.' && (p[2] == '\0' || p[2] == '/'))
+	/* Now look for pathname level-ups.  */
+	if (p[0] == '.' && p[1] == '.' && (p[2] == '\0' || ISDIRSEP (p[2])))
 	{
 	    --level;
 	    if (-level > max_level)
 		max_level = -level;
 	}
-	else if (p[0] == '\0' || p[0] == '/' ||
-		 (p[0] == '.' && (p[1] == '\0' || p[1] == '/')))
+	else if (p[0] == '\0' || ISDIRSEP (p[0]) ||
+		 (p[0] == '.' && (p[1] == '\0' || ISDIRSEP (p[1]))))
 	    ;
 	else
 	    ++level;
-	p = q;
-    } while (p != NULL);
+
+	/* q = strchr (p, '/'); but sub ISDIRSEP() for '/': */
+	while (*p != '\0' && !ISDIRSEP (*p)) p++;
+	if (*p != '\0') p++;
+    } while (*p != '\0');
     return max_level;
 }
 
-
+
+
 /* Free a vector, where (*ARGV)[0], (*ARGV)[1], ... (*ARGV)[*PARGC - 1]
    are malloc'd and so is *ARGV itself.  Such a vector is allocated by
    line2argv or expand_wild, for example.  */
@@ -283,13 +298,12 @@ compare_revnums (rev1, rev2)
     const char *rev1;
     const char *rev2;
 {
-    const char *s, *sp;
-    const char *t, *tp;
+    const char *sp, *tp;
     char *snext, *tnext;
     int result = 0;
 
-    sp = s = rev1;
-    tp = t = rev2;
+    sp = rev1;
+    tp = rev2;
     while (result == 0)
     {
 	result = strtoul (sp, &snext, 10) - strtoul (tp, &tnext, 10);
@@ -302,6 +316,9 @@ compare_revnums (rev1, rev2)
     return result;
 }
 
+/* Increment a revision number.  Working on the string is a bit awkward,
+   but it avoid problems with integer overflow should the revision numbers
+   get really big.  */
 char *
 increment_revnum (rev)
     const char *rev;
@@ -310,17 +327,29 @@ increment_revnum (rev)
     int lastfield;
     size_t len = strlen (rev);
 
-    newrev = (char *) xmalloc (len + 2);
+    newrev = xmalloc (len + 2);
     memcpy (newrev, rev, len + 1);
-    p = strrchr (newrev, '.');
-    if (p == NULL)
+    for (p = newrev + len; p != newrev; )
     {
-	free (newrev);
-	return NULL;
+	--p;
+	if (!isdigit(*p))
+	{
+	    ++p;
+	    break;
+	}
+	if (*p != '9')
+	{
+	    ++*p;
+	    return newrev;
+	}
+	*p = '0';
     }
-    lastfield = atoi (++p);
-    sprintf (p, "%d", lastfield + 1);
-
+    /* The number was all 9s, so change the first character to 1 and add
+       a 0 to the end.  */
+    *p = '1';
+    p = newrev + len;
+    *p++ = '0';
+    *p = '\0';
     return newrev;
 }
 
@@ -393,6 +422,71 @@ get_date (date, now)
 #endif
 #endif
 
+
+
+/* Given some revision, REV, return the first prior revision that exists in the
+ * RCS file, RCS.
+ *
+ * ASSUMPTIONS
+ *   REV exists.
+ *
+ * INPUTS
+ *   RCS	The RCS node pointer.
+ *   REV	An existing revision in the RCS file referred to by RCS.
+ *
+ * RETURNS
+ *   The first prior revision that exists in the RCS file, or NULL if no prior
+ *   revision exists.  The caller is responsible for disposing of this string.
+ *
+ * NOTES
+ *   This function currently neglects the case where we are on the trunk with
+ *   rev = X.1, where X != 1.  If rev = X.Y, where X != 1 and Y > 1, then this
+ *   function should work fine, as revision X.1 must exist, due to RCS rules.
+ */
+char *
+previous_rev (rcs, rev)
+    RCSNode *rcs;
+    const char *rev;
+{
+    char *p;
+    char *tmp = xstrdup (rev);
+    long r1;
+    char *retval;
+
+    /* Our retval can have no more digits and dots than our input revision.  */
+    retval = xmalloc (strlen (rev) + 1);
+    p = strrchr (tmp, '.');
+    *p = '\0';
+    r1 = strtol (p+1, NULL, 10);
+    do {
+	if (--r1 == 0)
+	{
+		/* If r1 == 0, then we must be on a branch and our parent must
+		 * exist, or we must be on the trunk with a REV like X.1.
+		 * We are neglecting the X.1 with X != 1 case by assuming that
+		 * there is no previous revision when we discover we were on
+		 * the trunk.
+		 */
+		p = strrchr (tmp, '.');
+		if (p == NULL)
+		    /* We are on the trunk.  */
+		    retval = NULL;
+		else
+		{
+		    *p = '\0';
+		    sprintf (retval, "%s", tmp);
+		}
+		break;
+	}
+	sprintf (retval, "%s.%ld", tmp, r1);
+    } while (!RCS_exist_rev (rcs, retval));
+
+    free (tmp);
+    return retval;
+}
+
+
+
 /* Given two revisions, find their greatest common ancestor.  If the
    two input revisions exist, then rcs guarantees that the gca will
    exist.  */
@@ -403,9 +497,9 @@ gca (rev1, rev2)
     const char *rev2;
 {
     int dots;
-    char *gca;
-    const char *p[2];
-    int j[2];
+    char *gca, *g;
+    const char *p1, *p2;
+    int r1, r2;
     char *retval;
 
     if (rev1 == NULL || rev2 == NULL)
@@ -417,52 +511,27 @@ gca (rev1, rev2)
     /* The greatest common ancestor will have no more dots, and numbers
        of digits for each component no greater than the arguments.  Therefore
        this string will be big enough.  */
-    gca = xmalloc (strlen (rev1) + strlen (rev2) + 100);
+    g = gca = xmalloc (strlen (rev1) + strlen (rev2) + 100);
 
     /* walk the strings, reading the common parts. */
-    gca[0] = '\0';
-    p[0] = rev1;
-    p[1] = rev2;
+    p1 = rev1;
+    p2 = rev2;
     do
     {
-	int i;
-	char c[2];
-	char *s[2];
-	
-	for (i = 0; i < 2; ++i)
-	{
-	    /* swap out the dot */
-	    s[i] = strchr (p[i], '.');
-	    if (s[i] != NULL) {
-		c[i] = *s[i];
-	    }
-	    
-	    /* read an int */
-	    j[i] = atoi (p[i]);
-	    
-	    /* swap back the dot... */
-	    if (s[i] != NULL) {
-		*s[i] = c[i];
-		p[i] = s[i] + 1;
-	    }
-	    else
-	    {
-		/* or mark us at the end */
-		p[i] = NULL;
-	    }
-	    
-	}
+	r1 = strtol (p1, (char **) &p1, 10);
+	r2 = strtol (p2, (char **) &p2, 10);
 	
 	/* use the lowest. */
-	(void) sprintf (gca + strlen (gca), "%d.",
-			j[0] < j[1] ? j[0] : j[1]);
+	(void) sprintf (g, "%d.", r1 < r2 ? r1 : r2);
+	g += strlen (g);
+	if (*p1 == '.') ++p1;
+	else break;
+	if (*p2 == '.') ++p2;
+	else break;
+    } while (r1 == r2);
 
-    } while (j[0] == j[1]
-	     && p[0] != NULL
-	     && p[1] != NULL);
-
-    /* back up over that last dot. */
-    gca[strlen(gca) - 1] = '\0';
+    /* erase that last dot. */
+    *--g = '\0';
 
     /* numbers differ, or we ran out of strings.  we're done with the
        common parts.  */
@@ -472,12 +541,8 @@ gca (rev1, rev2)
     {
 	/* revisions differ in trunk major number.  */
 
-	char *q;
-	const char *s;
-
-	s = (j[0] < j[1]) ? p[0] : p[1];
-
-	if (s == NULL)
+	if (r2 < r1) p1 = p2;
+	if (*p1 == '\0')
 	{
 	    /* we only got one number.  this is strange.  */
 	    error (0, 0, "bad revisions %s or %s", rev1, rev2);
@@ -486,13 +551,10 @@ gca (rev1, rev2)
 	else
 	{
 	    /* we have a minor number.  use it.  */
-	    q = gca + strlen (gca);
-	    
-	    *q++ = '.';
-	    for ( ; *s != '.' && *s != '\0'; )
-		*q++ = *s++;
-	    
-	    *q = '\0';
+	    *g++ = '.';
+	    while (*p1 != '.' && *p1 != '\0')
+		*g++ = *p1++;
+	    *g = '\0';
 	}
     }
     else if ((dots & 1) == 0)
@@ -500,10 +562,8 @@ gca (rev1, rev2)
 	/* if we have an even number of dots, then we have a branch.
 	   remove the last number in order to make it a revision.  */
 	
-	char *s;
-
-	s = strrchr(gca, '.');
-	*s = '\0';
+	g = strrchr (gca, '.');
+	*g = '\0';
     }
 
     retval = xstrdup (gca);
@@ -552,9 +612,10 @@ check_numeric (rev, argc, argv)
  */
 char *
 make_message_rcslegal (message)
-     char *message;
+     const char *message;
 {
-    char *dst, *dp, *mp;
+    char *dst, *dp;
+    const char *mp;
 
     if (message == NULL) message = "";
 
@@ -587,6 +648,61 @@ make_message_rcslegal (message)
     return dst;
 }
 
+
+
+/*
+ * file_has_conflict
+ *
+ * This function compares the timestamp of a file with ts_conflict set
+ * to the timestamp on the actual file and returns TRUE or FALSE based
+ * on the results.
+ *
+ * This function does not check for actual markers in the file and
+ * file_has_markers() function should be called when that is interesting.
+ *
+ * ASSUMPTIONS
+ *  The ts_conflict field is not NULL.
+ *
+ * RETURNS
+ *  TRUE	ts_conflict matches the current timestamp.
+ *  FALSE	The ts_conflict field does not match the file's
+ *		timestamp.
+ */
+int
+file_has_conflict (finfo, ts_conflict)
+    const struct file_info *finfo;
+    const char *ts_conflict;
+{
+    char *filestamp;
+    int retcode;
+
+    /* If ts_conflict is NULL, there was no merge since the last
+     * commit and there can be no conflict.
+     */
+    assert (ts_conflict);
+
+    /*
+     * If the timestamp has changed and no
+     * conflict indicators are found, it isn't a
+     * conflict any more.
+     */
+
+#ifdef SERVER_SUPPORT
+    if (server_active)
+	retcode = ts_conflict[0] == '=';
+    else 
+#endif /* SERVER_SUPPORT */
+    {
+	filestamp = time_stamp (finfo->file);
+	retcode = !strcmp (ts_conflict, filestamp);
+	free (filestamp);
+    }
+
+    return retcode;
+}
+
+
+
 /* Does the file FINFO contain conflict markers?  The whole concept
    of looking at the contents of the file to figure out whether there are
    unresolved conflicts is kind of bogus (people do want to manage files
@@ -729,19 +845,18 @@ void
 resolve_symlink (filename)
      char **filename;
 {
-    if ((! filename) || (! *filename))
+    if (filename == NULL || *filename == NULL)
 	return;
 
     while (islink (*filename))
     {
-	char *newname;
 #ifdef HAVE_READLINK
 	/* The clean thing to do is probably to have each filesubr.c
 	   implement this (with an error if not supported by the
 	   platform, in which case islink would presumably return 0).
 	   But that would require editing each filesubr.c and so the
 	   expedient hack seems to be looking at HAVE_READLINK.  */
-	newname = xreadlink (*filename);
+	char *newname = xreadlink (*filename);
 	
 	if (isabsolute (newname))
 	{
@@ -750,7 +865,7 @@ resolve_symlink (filename)
 	}
 	else
 	{
-	    char *oldname = last_component (*filename);
+	    const char *oldname = last_component (*filename);
 	    int dirlen = oldname - *filename;
 	    char *fullnewname = xmalloc (dirlen + strlen (newname) + 1);
 	    strncpy (fullnewname, *filename, dirlen);
@@ -831,6 +946,8 @@ shell_escape(buf, str)
     return buf;
 }
 
+
+
 /*
  * We can only travel forwards in time, not backwards.  :)
  */
@@ -883,7 +1000,8 @@ sleep_past (desttime)
 	    struct timeval tv;
 	    tv.tv_sec = s;
 	    tv.tv_usec = us;
-	    (void)select (0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &tv);
+	    (void)select (0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL,
+                          &tv);
 	}
 #else
 	if (us > 0) s++;
@@ -891,3 +1009,14 @@ sleep_past (desttime)
 #endif
     }
 }
+
+
+
+/* Return non-zero iff FILENAME is absolute.
+   Trivial under Unix, but more complicated under other systems.  */
+int
+isabsolute (filename)
+    const char *filename;
+{
+    return ISABSOLUTE (filename);
+}
